/*
** Programmer: Ciaran McHale (cjmchale@dsg.cs.tcd.ie)
**
** Routines to parse and process a pal configuration file.
** Parsing is performed by using recursive descent.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "globals.h"


#define	EOF_SYM		1
#define	DIMEN_VAR_SYM	2
#define	STRING_VAR_SYM	3
#define	REAL_SYM	4
#define	INT_VAR_SYM	5
#define	STRING_SYM	6
#define	EQUALS_SYM	7
#define	INT_SYM		8
#define	UNITS_SYM	9
#define	UNKNOWN_SYM	10

static struct {
	int	type;
	char	spelling[256];
	double	real_val;
	long	int_val;
	double	units_scale_factor;
	int	location;
} token;		/* lookahead token used in parser */


static struct {
	char	*spelling;
	double	*var;		/* pointer to relevant variable */
	BOOL	is_set;
} dimen_vars[] = {
	{ "PageWidth",			   &g.config.page_width,	FALSE },
	{ "PageHeight",			   &g.config.page_height,	FALSE },
	{ "LeftSideMargin",		   &g.config.left_margin,	FALSE },
	{ "TopMargin",			   &g.config.top_margin,	FALSE },
	{ "LabelWidth",			   &g.config.label_width,	FALSE },
	{ "LabelHeight",		   &g.config.label_height,	FALSE },
	{ "HorizontalSpaceBetweenLabels",  &g.config.h_gap,		FALSE },
	{ "VerticalSpaceBetweenLabels",	   &g.config.v_gap,		FALSE },
	{ "HorizontalMarginInsideLabels",  &g.config.h_inside_margin,	FALSE },
	{ "VerticalMarginInsideLabels",	   &g.config.v_inside_margin,	FALSE },
};


static struct {
	char	*spelling;
	int	*var;		/* pointer to relevant variable */
	BOOL	is_set;
} int_vars[] = {
	{ "NumberOfRows",	&g.config.num_rows,			FALSE },
	{ "NumberOfColumns",	&g.config.num_cols,			FALSE },
	{ "MaximumFontSize",	&g.config.max_font_size,		FALSE },
};


static struct {
	char	*spelling;
	char	**var;		/* pointer to relevant variable */
	BOOL	is_set;
} string_vars[] = {
	{ "FontFileName",	&g.config.font_file,			FALSE },
};


static struct {
	char	*spelling;
	double	scale_factor;
} units[] = {
	{ "in", 72.0 },
	{ "cm", 28.346457 },
	{ "mm", 2.8346457 },
};


static void parse_stmt(void);
static void parse_dimen_assignment(void);
static void parse_string_assignment(void);
static void parse_int_assignment(void);
static void accept(int token_type, char *err_msg);

static void next_token(void);
static void syntax_error(char *msg);
static void next_char(void);
static void init_lexical_analyser(FILE *file);
static void check_reserved_words(void);


void
parse_config_file(FILE *cfile)
{
	init_lexical_analyser(cfile);
	next_token();
	do {
		parse_stmt();
	} while (token.type != EOF_SYM);
	fclose(cfile);
}


static void
parse_stmt(void)
{
	switch (token.type) {
	case DIMEN_VAR_SYM:
		parse_dimen_assignment();
		break;
	case STRING_VAR_SYM:
		parse_string_assignment();
		break;
	case INT_VAR_SYM:
		parse_int_assignment();
		break;
	case STRING_SYM:
		syntax_error("Expecting a variable name for an "
				"assignment statement.\n"
				"Misspelt variable name perhaps?");
	default:
		syntax_error("Expecting a variable name for an "
				"assignment statement.");
	}
}


static void
parse_dimen_assignment(void)
{
	int	var_loc;
	double	var_value;

	var_loc = token.location;
	next_token();
	accept(EQUALS_SYM, "Expecting `='");
	switch (token.type) {
	case REAL_SYM:
		var_value = token.real_val;
		next_token();
		break;
	case INT_SYM:
		var_value = 1.0 * token.int_val;
		next_token();
		break;
	default:
		syntax_error("Expecting a number");
	}
	if (token.type == UNITS_SYM) {
		var_value *= token.units_scale_factor;
		next_token();
		*(dimen_vars[var_loc].var) = var_value;
		dimen_vars[var_loc].is_set = TRUE;
	} else {
		syntax_error("Expecting a unit (`in', `cm', `mm' or `pt')");
	}
}


static void
parse_string_assignment(void)
{
	int	var_loc;

	var_loc = token.location;
	next_token();
	accept(EQUALS_SYM, "Expecting `='");
	if (token.type == STRING_SYM) {
		if (string_vars[var_loc].is_set) {
			free( *(string_vars[var_loc].var) );
		}
		*(string_vars[var_loc].var) = str_malloc(token.spelling);
		string_vars[var_loc].is_set = TRUE;
		next_token();
	} else {
		syntax_error("Expecting a string");
	}
}


static void
parse_int_assignment(void)
{
	int	var_loc;

	var_loc = token.location;
	next_token();
	accept(EQUALS_SYM, "Expecting `='");
	if (token.type == INT_SYM) {
		*(int_vars[var_loc].var) = token.int_val;
		int_vars[var_loc].is_set = TRUE;
		next_token();
	} else {
		syntax_error("Expecting an integer");
	}
}


static void
accept(int token_type, char *err_msg)
{
	if (token.type == token_type) {
		next_token();
	} else {
		syntax_error(err_msg);
	}
}


static int	ch;	/* lookahead character used  in lexical analyser; */
static FILE	*config_file;	/* file that lexical analyser reads from */
static int	line_num;


static void
init_lexical_analyser(FILE *file)
{
	config_file = file;
	line_num = 1;
	ch = fgetc(config_file);
}


static void
next_char(void)
{
	ch = fgetc(config_file);
	switch (ch) {
	case '\n':
		line_num ++;
		break;
	case EOF:
		break;
	default:
		break;
	}
}


static void
syntax_error(char *msg)
{
	fprintf(stderr, "Error in configuration file on line %d "
			"near `%s'\n%s\n",
			line_num,
			token.spelling,
			msg);
	exit(1);
}


static void
next_token(void)
{
	/* skip white space */
	while (isspace(ch)) {
		next_char();
	}
	if (isalpha(ch)) {
		int	i;

		i = 0;
		while (!isspace(ch) && ch != '%') {
			token.spelling[i] = ch;
			i++;
			next_char();
		}
		token.spelling[i] = '\0';
		check_reserved_words();
		return;
	}
	if (isdigit(ch)) {
		int	i;
		int	num_dots;

		i = 0; num_dots = 0;
		while (isdigit(ch) || ch == '.') {
			num_dots += (ch == '.');
			token.spelling[i] = ch;
			i++;
			next_char();
		}
		token.spelling[i] = '\0';
		if (num_dots == 0) {
			token.type = INT_SYM;
			sscanf(token.spelling, "%d", &token.int_val);
		} else if (num_dots == 1) {
			token.type = REAL_SYM;
			sscanf(token.spelling, "%lf", &token.real_val);
		} else {
			token.type = UNKNOWN_SYM;
		}
		return;
	}
	switch (ch) {
	case '=':
		next_char();
		strcpy(token.spelling, "=");
		token.type = EQUALS_SYM;
		return;
	case '%':
		/* comment; consume it and start again */
		while (ch != '\n') {
			next_char();
		}
		next_token();
		return;
	case EOF:
		token.type = EOF_SYM;
		strcpy(token.spelling, "<End of file>");
		return;
	default:
		token.spelling[0] = ch;
		token.spelling[1] = '\0';
		token.type = UNKNOWN_SYM;
		return;
	}
}


static void check_reserved_words(void)
{
	int	i;

	for (i = 0; i < sizeof(units)/sizeof(units[0]); i++) {
		if (!strcmp(token.spelling, units[i].spelling)) {
			token.units_scale_factor = units[i].scale_factor;
			token.type = UNITS_SYM;
			return;
		}
	}
	for (i = 0; i < sizeof(dimen_vars)/sizeof(dimen_vars[0]); i++) {
		if (!strcmp(token.spelling, dimen_vars[i].spelling)) {
			token.location = i;
			token.type = DIMEN_VAR_SYM;
			return;
		}
	}
	for (i = 0; i < sizeof(int_vars)/sizeof(int_vars[0]); i++) {
		if (!strcmp(token.spelling, int_vars[i].spelling)) {
			token.location = i;
			token.type = INT_VAR_SYM;
			return;
		}
	}
	for (i = 0; i < sizeof(string_vars)/sizeof(string_vars[0]); i++) {
		if (!strcmp(token.spelling, string_vars[i].spelling)) {
			token.location = i;
			token.type = STRING_VAR_SYM;
			return;
		}
	}
	token.type = STRING_SYM;
}


void print_config_file_values(void)
{
	int	i;

	for (i = 0; i < sizeof(dimen_vars)/sizeof(dimen_vars[0]); i++) {
		printf("`%s' = %f\n",
			dimen_vars[i].spelling,
			*dimen_vars[i].var);
	}
	for (i = 0; i < sizeof(int_vars)/sizeof(int_vars[0]); i++) {
		printf("`%s' = %d\n",
			int_vars[i].spelling,
			*int_vars[i].var);
	}
	for (i = 0; i < sizeof(string_vars)/sizeof(string_vars[0]); i++) {
		printf("`%s' = %s\n",
			string_vars[i].spelling,
			*string_vars[i].var);
	}
}


void check_config_file_values(void)
{
	int	i;

	for (i = 0; i < sizeof(dimen_vars)/sizeof(dimen_vars[0]); i++) {
		if (!dimen_vars[i].is_set) {
			fprintf(stderr, "%s: Somebody forgot to set %s %s\n",
				g.prog_name, dimen_vars[i].spelling,
				"in the config file");
			exit(1);
		}
	}
	for (i = 0; i < sizeof(int_vars)/sizeof(int_vars[0]); i++) {
		if (!int_vars[i].is_set) {
			fprintf(stderr, "%s: Somebody forgot to set %s %s\n",
				g.prog_name, int_vars[i].spelling,
				"in the config file");
			exit(1);
		}
	}
	for (i = 0; i < sizeof(string_vars)/sizeof(string_vars[0]); i++) {
		if (!string_vars[i].is_set) {
			fprintf(stderr, "%s: Somebody forgot to set %s %s\n",
				g.prog_name, string_vars[i].spelling,
				"in the config file");
			exit(1);
		}
	}
}
