/*
** Programmer: Ciaran McHale (cjmchale@dsg.cs.tcd.ie)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "globals.h"

struct GlobalsType	g;


static void	usage();
static void	process_cmd_line_options(int argc, char *argv[],
					 char **cfile_name);
static void	find_and_process_afm_file();
static void	start_ps_document(PS_generator psg, FILE *font_file);
static void	end_ps_document(PS_generator psg, int num_pages);
static void	determine_input_and_output(int argc, char *argv[], FILE **ifile,
					 PS_generator *psg);


int
main(int argc, char *argv[])
{
	char		*cfile_name;
	FILE		*cfile, *ifile, *font_file;
	BOOL		is_pfa_file;
	PS_generator	psg;
	int		num_pages;

	g.prog_name = argv[0];

	/*
	** Process command line options
	*/
	process_cmd_line_options(argc, argv, &cfile_name);

	/*
	** Process the configuration file
	*/
	cfile = open_lib_file(cfile_name, "configuration file");
	parse_config_file(cfile);
	fclose(cfile);
	check_config_file_values();

	/*
	** The configuration file specified, among other things, which
	** font should be used. Find the Adode Font Metrics (AFM) file
	** for this font and process it.
	*/
	find_and_process_afm_file();

	/*
	** Look at any remaining command line parameters to determine
	** where input is to be read from (default is stdin) and where
	** output is to be written (default is stdout).
	*/
	determine_input_and_output(argc, argv, &ifile, &psg);

	/*
	** Write the header and prolog of the output file
	** and then create the "Setup" section of the output file
	*/
	if (g.include_font) {
		/*
		** Try to locate & open the .pfa file for the font.
		** If that fails then try for the .pfb file.
		*/
		font_file = open_font_file(g.config.font_file, "pfa");
		if (font_file) {
			is_pfa_file = TRUE;
		} else {
			is_pfa_file = FALSE;
			font_file = open_font_file(g.config.font_file, "pfb");
		}
	} else {
		font_file = NULL;
	}
	start_ps_document(psg, font_file);
	generate_setup_section(psg, font_file, is_pfa_file);
	if (font_file) {
		fclose(font_file);
	}

	/*
	** Create the main body of the output file
	*/
	g.doc_psbbox = make_empty_psbbox();
	if (g.page_per_label) {
		process_input_page_per_label(psg, ifile, &num_pages);
	} else {
		process_input_normally(psg, ifile, &num_pages);
	}
	fclose(ifile);

	/*
	** Write the trailer to the output file
	*/
	end_ps_document(psg, num_pages);

	/*
	** Perform some tidying up before exiting.
	** This isn't really necessary on UNIX but what the heck.
	*/
	ps_gen_destroy(psg);
	destroy_my_FontInfo(&g.f, FALSE);
	return 0;
}


static void
usage()
{
	fprintf(stderr,
		"usage:\t%s [ options ] [ input-file [ output-file ] ]\n",
		g.prog_name);
	fprintf(stderr, "Options available:\n%s\n%s\n%s\n%s\n%s\n%s\n",
		"\t-c<config-file>",
		"\t-h\t\tProvides this help message",
		"\t-d<thickness>\tDraw box around labels",
		"\t-p\t\tOutputs a page of labels for each one specified",
		"\t-f\t\tDo not download the font program to the PS file",
		"\t-q\t\tQuiet mode (don't give warnings)"
		);
	exit(1);
}


static void
process_cmd_line_options(int argc, char *argv[], char **cfile_name)
{
	int	ch;

	/*
	** Set some defaults
	*/
	*cfile_name = getenv("PAL_CONFIG_FILE");
	if ( !(*cfile_name) ) {
		*cfile_name = DEFAULT_CONFIG_FILE;
	}
	g.draw_box = FALSE;
	g.page_per_label = FALSE;
	g.include_font = TRUE;
	g.quiet_mode = FALSE;

	/*
	** Now process the command line options which may overide the
	** above defaults
	*/
	while ((ch = getopt(argc, argv, "d:pc:fqh")) != EOF) {
		switch (ch) {
		case 'd':
			g.draw_box = TRUE;
			if (sscanf(optarg, "%lf", &g.box_thickness) != 1) {
				usage();
			}
			break;
		case 'p':
			g.page_per_label = TRUE;
			break;
		case 'c':
			*cfile_name = optarg;
			break;
		case 'f':
			g.include_font = FALSE;
			break;
		case 'q':
			g.quiet_mode = TRUE;
			break;
		case 'h':
		default:
			usage(argv[0]);		/* doesn't return */
		}
	}
}


static void
determine_input_and_output(int argc, char *argv[], FILE **ifile,
			   PS_generator *psg)
{
	switch (argc-optind) {
	case 0:
		/*
		** No input or output filenames specified.
		** Default to stdin/stdout.
		*/
		*ifile = stdin;
		*psg = make_PS_generator(NULL);
		break;
	case 1:
		*ifile = fopen(argv[optind], "r");
		if ( !(*ifile) ) {
			fprintf(stderr, "%s: cannot open `%s'\n",
				argv[0], argv[optind]);
			exit(1);
		}
		*psg = make_PS_generator(NULL);
		break;
	case 2:
		*ifile = fopen(argv[optind], "r");
		if ( !(*ifile) ) {
			fprintf(stderr, "%s: cannot open `%s'\n",
				g.prog_name, argv[optind]);
			exit(1);
		}
		*psg = make_PS_generator(argv[optind+1]);
		if ( !(*psg) ) {
			fprintf(stderr, "%s: cannot open `%s'\n",
				g.prog_name, argv[optind+1]);
			exit(1);
		}
		break;
	default:
		/*
		** Incorrect number of command line parameters
		*/
		usage(argv[0]);			/* doesn't return */
	}
}


static void
find_and_process_afm_file()
{
	FILE		*afm_file;

	afm_file = open_font_file(g.config.font_file, "afm");
	if (!afm_file) {
	fprintf(stderr, "%s: can't find font metrics file `%s.afm'\n",
		g.prog_name, g.config.font_file);
		exit(1);
	}
	switch (parseFile(afm_file, &g.f.f, P_ALL)) {
	case parseError:
		fprintf(stderr, "%s: error in parsing file `%s.afm'\n",
			g.prog_name, g.config.font_file);
		exit(1);
	case ok:
		fclose(afm_file);
		break;
	case earlyEOF:
		fprintf(stderr, "%s: the file `%s.afm' ended early\n",
			g.prog_name, g.config.font_file);
		exit(1);
	case storageProblem:
		fprintf(stderr, "%s: out of memory\n", g.prog_name);
		exit(1);
	default:
		fprintf(stderr, "%s: unidentified error while %s `%s.afm'\n",
			g.prog_name,
			"parsing the file",
			g.config.font_file);
		exit(1);
	}
	check_and_reencode_font(&g.f);
}


static void	
start_ps_document(PS_generator psg, FILE *font_file)
{
	char		str[256];
	FILE		*pfile;

	ps_gen_DSC_comment(psg, "%!PS-Adobe-3.0");
	sprintf(str, "%s%s", "%%Creator: ", g.prog_name);
	ps_gen_DSC_comment(psg, str);
	ps_gen_DSC_comment(psg, "%%Pages: (atend)");
	ps_gen_DSC_comment(psg, "%%BoundingBox: (atend)");
	ps_gen_DSC_comment(psg, "%%LanguageLevel: 1");
	ps_gen_DSC_comment(psg, "%%Orientation: Portrait");
	ps_gen_DSC_comment(psg, "%%PageOrder: Ascend");
	if (!font_file) {
		sprintf(str, "%s%s", "%%DocumentNeededResources: font ",
			g.f.f->gfi->fontName);
	} else {
		sprintf(str, "%s%s", "%%DocumentSuppliedResources: font ",
			g.f.f->gfi->fontName);
	}
	ps_gen_DSC_comment(psg, str);
	ps_gen_DSC_comment(psg, "%%EndComments");
	
	/*
	** Write the prolog to the output file
	*/
	pfile = open_lib_file("prolog.ps", "library file");
	ps_gen_append_ascii_file(psg, pfile);
	fclose(pfile);
}


static void	
end_ps_document(PS_generator psg, int num_pages)
{
	char		str[256];
	double		x1, y1, x2, y2;

	ps_gen_DSC_comment(psg, "%%Trailer");
	ps_gen_command(psg, "end");
	ps_gen_comment(psg, "% pal_dict");
	sprintf(str, "%s%d", "%%Pages: ", num_pages);
	ps_gen_DSC_comment(psg, str);
	/* %%BoundingBox */
	psbbox_get_corners(g.doc_psbbox, &x1, &y1, &x2, &y2);
	sprintf(str, "%s %d %d %d %d",
		"%%BoundingBox:",
		(int)floor(x1), (int)floor(y1), (int)ceil(x2), (int)ceil(y2)
		);
	ps_gen_DSC_comment(psg, str);
	/* %%EOF */
	ps_gen_DSC_comment(psg, "%%EOF");
}
