/*
** Programmer: Ciaran McHale (cjmchale@dsg.cs.tcd.ie)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "globals.h"

static LabelInfo *read_label(FILE *file);
static char	 *read_input_line(FILE *file);
static LabelInfo *make_new_label(char *line);
static void	 delete_label(LabelInfo *);
static void	 add_line_to_label(LabelInfo *l, char *line);
static void	 copy_with_translation(char *out, char *in);
static void	 start_ps_page(PS_generator psg, int page_num);
static void	 end_ps_page(PS_generator psg);


void
process_input_normally(PS_generator psg, FILE *ifile, int *num_pages)
{
	LabelInfo	*label;
	int		num_labels;
	const int	num_labels_per_page =
					g.config.num_rows * g.config.num_cols;

	*num_pages = 0;
	num_labels = 0;
	while ((label = read_label(ifile)) != NULL) {
		process_label(label);
		if (label->okay) {
			num_labels ++;
			if (num_labels == 1) {
				*num_pages += 1;
				start_ps_page(psg, *num_pages);
			}
			print_label(psg, label, num_labels);
		} else if (!g.quiet_mode) {
			fprintf(stderr, "Ignoring label addressed to ");
			if (strlen(label->lines[0]) > 23) {
				strcpy(&(label->lines[0][19]), "...");
			}
			fprintf(stderr, "`%s': as it is too big\n",
				label->lines[0]);
		}

		if (num_labels == num_labels_per_page) {
			end_ps_page(psg);
			num_labels = 0;
		}
		delete_label(label);
	}
	if (num_labels != 0) {
		end_ps_page(psg);
	}
}


void
process_input_page_per_label(PS_generator psg, FILE *ifile, int *num_pages)
{
	LabelInfo	*label;
	const int	num_labels_per_page =
					g.config.num_rows * g.config.num_cols;
	int		i;

	*num_pages = 0;
	while ((label = read_label(ifile)) != NULL) {
		process_label(label);
		if (label->okay) {
			*num_pages += 1;
			start_ps_page(psg, *num_pages);

			for (i = 0; i < num_labels_per_page; i++) {
				print_label(psg, label, i + 1);
			}
			end_ps_page(psg);
		} else if (!g.quiet_mode) {
			fprintf(stderr, "Ignoring label addressed to ");
			if (strlen(label->lines[0]) > 23) {
				strcpy(&(label->lines[0][19]), "...");
			}
			fprintf(stderr, "`%s': as it is too big\n",
				label->lines[0]);
		}
		delete_label(label);
	}
}


static void
start_ps_page(PS_generator psg, int page_num)
{
	char	str[256];

	g.page_psbbox = make_empty_psbbox();

	sprintf(str, "%s%d %d", "%%Page: ", page_num, page_num);
	ps_gen_DSC_comment(psg, str);
	ps_gen_DSC_comment(psg, "%%PageBoundingBox: (atend)");
	ps_gen_DSC_comment(psg, "%%BeginPageSetup");
	ps_gen_command(psg, "pi");
	ps_gen_DSC_comment(psg, "%%EndPageSetup");
}


static void
end_ps_page(PS_generator psg)
{
	char	str[256];
	double	x1, y1, x2, y2;

	ps_gen_command(psg, "pf");
	psbbox_union(g.doc_psbbox, g.page_psbbox, g.doc_psbbox);
	psbbox_get_corners(g.page_psbbox, &x1, &y1, &x2, &y2);
	destroy_psbbox(g.page_psbbox);
	ps_gen_DSC_comment(psg, "%%PageTrailer");
	sprintf(str, "%s %d %d %d %d",
		"%%PageBoundingBox:",
		(int)floor(x1), (int)floor(y1), (int)ceil(x2), (int)ceil(y2)
		);
	ps_gen_DSC_comment(psg, str);
}


/*
** Copy from "in" to "out", but translate escape sequences into byte codes.
*/
static void
copy_with_translation(char *out, char *in)
{
	int		i, j;
	char	ch;

	i = j = 0;
	while (in[i] != '\0') {
		switch (in[i]) {
		default:
			/*
			** The most common case---a non-special character
			*/
			out[j++] = in[i++];
			break;
		case '\\':
			i++;
			switch (in[i]) {
			case '%':
				out[j++] = in[i++];
				break;
			case '\'':
			case '`':
			case 'v':
			case ',':
			case '^':
			case '"':
			case 'o':
			case '~':
				ch = find_accented_char(in[i], in[i+1]);
				if (ch != 0) {
					out[j++] = ch;
				} else if (!g.quiet_mode) {
					/*
					** Accented char is not available.
					** Use non-accented version instead.
					*/
					fprintf(stderr,
						"Warning: %s `\\%c%c' %s\n",
						"Accented character",
						in[i], in[i+1],
						"not available."
						);
					out[j++] = in[i+1];
				}
				i += 2;
				break;
			default:
				if (isdigit(in[i]) && isdigit(in[i+1])
				    && isdigit(in[i+2])) {
					int	num;
					char	ch;

					ch = in[i+3];
					in[i+3] = '\0';
					sscanf(&in[i], "%o", &num);
					if (num == 0) {
						/*
						** A null character would
						** terminate the string
						** so we'll ignore it.
						*/
					} else {
						in[i+3] = ch;
						out[j++] = num;
					}
				} else {
					out[j++] = in[i++];
				}
				break;
			}
		}
	}
	out[j] = '\0';
}


static char *
read_input_line(FILE *file)
{
static	char	line[MAX_LEN];
	char	*p;
	int	i;

	while (1) {
		if ((p = fgets(line, MAX_LEN, file)) == NULL) {
			/* end of file */
			return NULL;
		}
		if (line[strlen(line)-1] == '\n') {
			line[strlen(line)-1] = '\0'; /* remove '\n' */
		} else {
			fprintf(stderr, "%s: Error: line in input file %s\n",
				g.prog_name, "too long");
			exit(1);
		}
		if (line[0] == '%')
			continue; /* the whole line is a comment; ignore it */
		/*
		** The line contains some text; however if might also
		** contain a comment; if so truncate the line to remove it
		*/
		for (i = 1; line[i] != '\0'; i++) {
			if (line[i] == '%' && line[i-1] != '\\') {
				/* comment extends to end of line, so... */
				line[i] = '\0';
				return line;
			}
		}
		return line;
	}
	fprintf(stderr, "%s: BUG in read_input_line(): %s\n",
		g.prog_name, "unreachable code reached");
	exit(1);
}


static LabelInfo *
read_label(FILE *file)
{
	char		*p;
	LabelInfo	*l;

	/*
	** Skip blank lines before the label
	*/
	while ((p = read_input_line(file)) != NULL && p[0] == '\0')
		;

	if (p == NULL) {
		/* end of file so no label to return */
		return (LabelInfo *)0;
	}
	/*
	** We got the first line of the label
	*/
	l = make_new_label(p);
	while ((p = read_input_line(file)) != NULL && p[0] != '\0') {
		add_line_to_label(l, p);
	}
	return l;
}


static LabelInfo *
make_new_label(char *line)
{
	LabelInfo	*l;

	l = (LabelInfo *)my_malloc(sizeof(LabelInfo));
	l->font_size = 0;
	l->width = l->height = 0.0;
	l->okay = TRUE;
	l->max_lines = 10;
	l->lines = (char **)my_malloc( sizeof(char *) * l->max_lines );
	l->num_lines = 0;
	add_line_to_label(l, line);
	return l;
}


static void
delete_label(LabelInfo *label)
{
	int	i;

	for (i = 0; i < label->num_lines; i++) {
		free(label->lines[i]);
	}
	free(label->lines);
	free(label);
}


static void
add_line_to_label(LabelInfo *l, char *line)
{
	if (l->num_lines == l->max_lines) {
		/* need to make room for more lines */
		l->max_lines += 10;
		l->lines = (void *)my_realloc(l->lines,
						sizeof(char *) * l->max_lines);
	}
	l->lines[l->num_lines] = str_malloc(line);
	copy_with_translation(l->lines[l->num_lines], line);
	l->num_lines ++;
}
