/*
** Programmer: Ciaran McHale (cjmchale@dsg.cs.tcd.ie)
**
** Routines to process a single address label
*/

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include "globals.h"


static int vert_font_size(LabelInfo *label, int font_size, double max_height);
static int horiz_font_size(LabelInfo *label, int font_size, double max_width);
static void max_line_width(LabelInfo *label, int font_size,
		double *stringwidth, double *left_pt, double *right_pt);
static void line_width(char *str, long *stringwidth, long *l_pt, long *r_pt);
static double str_descent(char *str, int font_size);
static double str_ascent_over_capheight(char *str, int font_size);

void
process_label(LabelInfo *label)
{
	int	font_size;
	double	internal_width, internal_height;

	internal_height = g.config.label_height - 2 * g.config.v_inside_margin;
	internal_width = g.config.label_width - 2 * g.config.h_inside_margin;
	font_size = g.config.max_font_size;
	font_size = vert_font_size(label, font_size, internal_height);
	font_size = horiz_font_size(label, font_size, internal_width);
	if (font_size == 0) {
		label->okay = FALSE;
	} else {
		label->okay = TRUE;
		label->font_size = font_size;
		max_line_width(label, font_size, &label->width, (double *)0,
				(double *)0);
		label->height = label_center_height(label, font_size);
	}
}


static int
vert_font_size(LabelInfo *label, int font_size, double max_height)
{
	double	scale, center_height, gap;

	/*
	** Estimate the amount of vertical space the label requires
	*/
	center_height = label_center_height(label, font_size);
	scale = max_height / center_height;
	if (scale < 1.0) {
		font_size = (int) floor( scale * (double)font_size );
		center_height = label_center_height(label, font_size);
	}
	assert(max_height >= center_height);

	/*
	** Vertically centered thus, will any descenders on the last line
	** of text dip down too low, or the anything on first line stretch
	** above the capheight too much?
	*/
	gap = (max_height - center_height) / 2.0;
	if (gap < -str_descent(label->lines[label->num_lines-1], font_size)
	    || gap < str_ascent_over_capheight(label->lines[0], font_size)){
		/*
		** The label won't fit at this font size so guess
		** a font size which is (hopefully) small enough and
		** recurse to confirm the guess.
		*/
		font_size = vert_font_size(label, font_size - 1, max_height);
	}
	return font_size;
}


static double
str_descent(char *str, int font_size)
{
	int		max_descent;
	int		i;
	CharMetricInfo	*cm;
	double		res;

	max_descent = 0;
	for (i = 0; str[i] != '\0'; i++) {
		cm = afm_find_char_info_of_code(&g.f, str[i]);
		if (cm->charBBox.lly < max_descent) {
			max_descent = cm->charBBox.lly;
		}
	}
	res = max_descent * font_size / 1000.0;
	return res;
}


static double
str_ascent(char *str, int font_size)
{
	int		max_ascent;
	int		i;
	CharMetricInfo	*cm;
	double		res;

	max_ascent = 0;
	for (i = 0; str[i] != '\0'; i++) {
		cm = afm_find_char_info_of_code(&g.f, str[i]);
		if (cm->charBBox.ury > max_ascent) {
			max_ascent = cm->charBBox.ury;
		}
	}
	res = (max_ascent) * font_size / 1000.0;
	return res;
}


static double
str_ascent_over_capheight(char *str, int font_size)
{
	int		max_ascent;
	int		i;
	CharMetricInfo	*cm;
	double		res;

	max_ascent = 0;
	for (i = 0; str[i] != '\0'; i++) {
		cm = afm_find_char_info_of_code(&g.f, str[i]);
		if (cm->charBBox.ury > max_ascent) {
			max_ascent = cm->charBBox.ury;
		}
	}
	res = (max_ascent - g.f.f->gfi->capHeight) * font_size / 1000.0;
	return res;
}


/*
** Calculate how much space the label would require if the text
** on the top line did not extend above the font's capHeight and the
** text on the last line did not have any descenders. This is the
** "height" that will be used for centering the text vertically.
*/
double
label_center_height(LabelInfo *label, int font_size)
{
	return  (label->num_lines - 1) * font_spacing(font_size)
		+ afm_capHeight(&g.f, font_size);
}


static int
horiz_font_size(LabelInfo *label, int font_size, double max_width)
{
	double	scale;
	double	stringwidth, left_pt, right_pt, text_origin;

	max_line_width(label, font_size, &stringwidth, &left_pt, &right_pt);
	if (stringwidth > max_width) {
		scale = max_width / stringwidth;
		font_size = (int) floor( scale * (double)font_size );
		stringwidth = max_width;
		left_pt = scale * left_pt;
		right_pt = scale * right_pt;
	}
	text_origin = (max_width - stringwidth) / 2.0;
	if (text_origin + left_pt < 0 || text_origin + right_pt > max_width) {
		font_size = horiz_font_size(label, font_size - 1, max_width);
	}
	return font_size;
}


static void
max_line_width(LabelInfo *label, int font_size,
		double *stringwidth, double *left_pt, double *right_pt)
{
	long	sw, lp, rp, max_sw, min_lp, max_rp;
	int	i;

	max_sw = min_lp = max_rp = 0;
	for (i = 0; i < label->num_lines; i++) {
		line_width(label->lines[i], &sw, &lp, &rp);
		max_sw = (sw > max_sw) ? sw : max_sw;
		min_lp = (lp < min_lp) ? lp : min_lp;
		max_rp = (rp > max_rp) ? rp : max_rp;
	}
	if (stringwidth)	*stringwidth = max_sw * font_size / 1000.0;
	if (left_pt)		*left_pt = min_lp * font_size / 1000.0;
	if (right_pt)		*right_pt = max_rp * font_size / 1000.0;
}


static void
line_width(char *str, long *stringwidth, long *left_pt, long *right_pt)
{
	int		i;
	long		width;
	CharMetricInfo	*c;

	width = 0;
	for (i = 0; str[i] != '\0'; i++) {
		width += afm_char_width(&g.f, str[i]);
	}
	if (i > 0) {
		if (stringwidth)	*stringwidth = width;

		c = afm_find_char_info_of_code(&g.f, str[0]);
		if (left_pt)	*left_pt = c->charBBox.llx;

		c = afm_find_char_info_of_code(&g.f, str[i-1]);
		if (right_pt)	*right_pt = width + (c->charBBox.urx - c->wx);
	}
}


double
font_spacing(int font_size)
{
	return 1.1 * font_size;
}


void
reg_font_size(int font_size)
{
	/* To do... */
}


void
print_label(PS_generator psg, LabelInfo *label, int label_num)
{
	char		str[256];
	double		label_x, label_y;
	double		text_x, text_y;
	int		col, row;
	int		i;
	PSBBox		label_psbbox;

	row = (label_num-1) % g.config.num_rows;
	col = (label_num-1) / g.config.num_rows;

	/*
	** Calculate position of top left corner of label
	*/
	label_x = g.config.left_margin +
				col * (g.config.label_width + g.config.h_gap);
	label_y = g.config.page_height - g.config.top_margin
			- row * (g.config.label_height + g.config.v_gap);

	/*
	** Calculate start position of first line of text
	*/
	text_x = label_x
		 + (g.config.label_width - label->width) / 2.0;
	text_y = label_y
		 - (g.config.label_height - label->height) / 2.0
		 - afm_capHeight(&g.f, label->font_size);

	if (g.draw_box) {
		double	stray;

		/*
		** psbbox is full size of label plus bit of line that
		** strays outside
		*/
		stray = g.box_thickness / 2.0;
		label_psbbox = make_psbbox(label_x - stray,
					label_y - g.config.label_height - stray,
					label_x + g.config.label_width + stray,
					label_y + stray
				      );

		/* generate PS code to actually draw the box */
		sprintf(str, "%.2f %.2f %.2f %.2f box stroke",
			label_x,
			label_y,
			label_x + g.config.label_width,
			label_y - g.config.label_height
			);
		ps_gen_command(psg, str);
	} else {
		double	x1, y1, x2, y2;
		double	left_most, right_most;
		int	font_size = label->font_size;

		/*
		** Have to calculate the precise psbbox of the text in the label
		*/
		y2 = text_y + str_ascent(label->lines[0], font_size);
		y1 = text_y + afm_capHeight(&g.f, font_size) - label->height
		     + str_descent(label->lines[label->num_lines-1], font_size);
		max_line_width(label, font_size, 0, &left_most, &right_most);
		x1 = text_x + left_most;
		x2 = text_x + right_most;
		label_psbbox = make_psbbox(x1, y1, x2, y2);
	}
	psbbox_union(g.page_psbbox, label_psbbox, g.page_psbbox);
	destroy_psbbox(label_psbbox);

	/*
	** Generate code to print the label
	*/
	i = g.config.max_font_size - label->font_size;
	if (i < 5) {
		sprintf(str, "f%d", i);
		ps_gen_command(psg, str);
	} else {
		sprintf(str, "%d f", label->font_size);
		ps_gen_command(psg, str);
	}
	ps_gen_newline(psg);
	for (i = 0; i < label->num_lines; i++) {
		sprintf(str, "%.2f %.2f m", text_x, text_y);
		ps_gen_command(psg, str);
		ps_gen_string(psg, label->lines[i]);
		ps_gen_command(psg, "s");
		ps_gen_newline(psg);
		text_y = text_y - font_spacing(label->font_size);
	}
	ps_gen_newline(psg);
}
