/*
** Programmer: Ciaran McHale (cjmchale@dsg.cs.tcd.ie)
**
** A data structure and routines to make it easier to generate
** PostScript code. In effect, it just encapsulates fprintf() calls but
** the PS_generator structure maintains some state to determine when
** linebreaking needs to be done, thus removing the burden of this from
** the programmer.
*/


#include <stdio.h>
#include "ps_gen.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>


extern	void	*my_malloc(int);


#define	MAX_LINE_LEN		72	/* to keep things tidy */
#define	PSG_NO_SPACE_NEEDED	1
#define	PSG_SPACE_NEEDED	2


struct _PS_generator {
	FILE	*file;
	int	col_num;
	int	state;
};

PS_generator
make_PS_generator(char *file_name)
{
	PS_generator	psg;

	psg = (PS_generator)my_malloc(sizeof(struct _PS_generator));
	if (!file_name) {
		psg->file = stdout;
	} else {
		psg->file = fopen(file_name, "w");
		if (!psg->file) {
			free(psg);
			return (PS_generator)0;
		}
	}
	psg->col_num = 0;
	psg->state = PSG_NO_SPACE_NEEDED;
	return psg;
}


void
ps_gen_destroy(PS_generator psg)
{
	fclose(psg->file);
	free(psg);
}


void
ps_gen_comment(PS_generator psg, char *str)
{
	int	len;

	len = strlen(str);
	assert(str[0] == '%' && str[len-1] != '\n');
	if (psg->col_num != 0 && psg->col_num+len > MAX_LINE_LEN) {
		ps_gen_newline(psg);
	}
	fprintf(psg->file, "%s\n", str);
	psg->col_num = 0;
	psg->state = PSG_NO_SPACE_NEEDED;
}


void
ps_gen_DSC_comment(PS_generator psg, char *str)
{
	assert(str[0]=='%' && (str[1]=='%' || str[1]=='!'));
	if (psg->col_num != 0) {
		ps_gen_newline(psg);
	}
	ps_gen_comment(psg, str);
}


void
ps_gen_command(PS_generator psg, char *str)
{
	int	len;

	len = strlen(str);
	assert(!isspace(str[0]) && !isspace(str[len-1]));
	if (len > MAX_LINE_LEN) {
		fprintf(stderr, "ps_gen_command(): command too long\n");
		exit(1);
	}
	if (psg->col_num + len > MAX_LINE_LEN) {
		/*
		** Put this command on a new line to prevent the current
		** line from getting too long
		*/
		fprintf(psg->file, "\n%s", str);
		psg->col_num = len;
	} else {
		switch (psg->state) {
		case PSG_NO_SPACE_NEEDED:
			/* no need to output a space before the command */
			fprintf(psg->file, "%s", str);
			psg->col_num += len;
			break;
		case PSG_SPACE_NEEDED:
			/* output a space before the command */
			fprintf(psg->file, " %s", str);
			psg->col_num += 1 + len;
			break;
		default:
			fprintf(stderr, "BUG in ps_gen_command()\n");
			exit(1);
		}
	}
	psg->state = PSG_SPACE_NEEDED;
}


void
ps_gen_string(PS_generator psg, char *str)
{
	int	len;
	int	i;

	len = strlen(str);
	if (psg->col_num + 5 >= MAX_LINE_LEN) {
		/*
		** Put this string on a new line
		*/
		fprintf(psg->file, "\n");
		psg->col_num = 0;
	}
	fprintf(psg->file, "(");
	psg->col_num += 1;
	for (i = 0; i < len; i++) {
		if (str[i] == '(' || str[i] == ')') {
			fprintf(psg->file, "\\%c", str[i]);
			psg->col_num += 2;
		} else if (32 <= str[i] && ((unsigned char *)str)[i] <= 126) {
			fprintf(psg->file, "%c", str[i]);
			psg->col_num += 1;
		} else {
			fprintf(psg->file, "\\%03o", ((unsigned char *)str)[i]);
			psg->col_num += 4;
		}
		if (psg->col_num >= MAX_LINE_LEN) {
			fprintf(psg->file, "\\\n");
			psg->col_num = 0;
		}
	}
	fprintf(psg->file, ")");
	psg->state = PSG_NO_SPACE_NEEDED;
}


void
ps_gen_newline(PS_generator psg)
{
	fprintf(psg->file, "\n");
	psg->col_num = 0;
	psg->state = PSG_NO_SPACE_NEEDED;
}


void
ps_gen_append_ascii_file(PS_generator psg, FILE *file)
{
	int	ch;
	int	prev_ch;

	if (psg->col_num != 0) {
		ps_gen_newline(psg);
	}
	prev_ch = '\n';
	while ((ch = getc(file)) != EOF) {
		putc(ch, psg->file);
		prev_ch = ch;
	}
	if (prev_ch != '\n') {
		ps_gen_newline(psg);
	}
}


void
ps_gen_append_lines(PS_generator psg, char *lines[], int num_lines)
{
	int	i, len;

	for (i = 0; i < num_lines; i++) {
		len = strlen(lines[i]);
		assert(lines[i][len-1] != '\n');
		fprintf(psg->file, "%s\n", lines[i]);
	}
}

/*
** The code used in the ps_gen_append_pfb_file() is a modification of
** that in the "t1ascii" utility program. Here is the copyright notice
** from that program.
*/

/* t1ascii
 *
 * This program takes an Adobe Type-1 font program in binary (PFB) format and
 * converts it to ASCII (PFA) format.
 *
 * Copyright (c) 1992 by I. Lee Hetherington, all rights reserved.
 *
 * Permission is hereby granted to use, modify, and distribute this program
 * for any purpose provided this copyright notice and the one below remain
 * intact. 
 *
 * I. Lee Hetherington (ilh@lcs.mit.edu)
 *
 * $Log:	t1ascii.c,v $
 * Revision 1.1  92/05/22  11:47:24  ilh
 * initial version
 * 
 */

static char copyright[] =
  "@(#) Copyright (c) 1992 by I. Lee Hetherington, all rights reserved.";

#define MARKER   128
#define ASCII    1
#define BINARY   2
#define DONE     3


/*
** This function reads a four-byte block length.
*/
static int read_length(FILE *file)
{
	int length;

	length = fgetc(file);
	length |= fgetc(file) << 8;
	length |= fgetc(file) << 16;
	length |= fgetc(file) << 24;
	return length;
}

/*
** This function outputs a single byte in hexadecimal.
** It limits hexadecimal output to 64 columns.
*/
static void output_hex(PS_generator psg, int b)
{
	static char *hexchar = "0123456789ABCDEF";
	static int hexcol = 0;

	/* trim hexadecimal lines to 64 columns */
	if (hexcol >= 64) {
		fputc('\n', psg->file);
		hexcol = 0;
	}
	fputc(hexchar[(b >> 4) & 0xf], psg->file);
	fputc(hexchar[b & 0xf], psg->file);
	hexcol += 2;
}


void
ps_gen_append_pfb_file(PS_generator psg, FILE *file)
{
	int	ch, length;
	int	last_type = ASCII;


	/*
	** main loop through blocks
	*/
	while (1) {
		ch = fgetc(file);
		if (ch == EOF) {
			break;
		}
		if (ch != MARKER) {
			fprintf(stderr, "error in PFB file\n");
			exit(1);
		}
		switch (ch = fgetc(file)) {
		case ASCII:
			if (last_type != ASCII) {
				fputc('\n', psg->file);
			}
			last_type = ASCII;
			for (length = read_length(file); length > 0; length--) {
				if ((ch = fgetc(file)) == '\r') {
					fputc('\n', psg->file);
				} else {
					fputc(ch, psg->file);
				}
			}
			break;
		case BINARY:
			last_type = BINARY;
			for (length = read_length(file); length > 0; length--) {
				output_hex(psg, fgetc(file));      
			}
			break;
		case DONE:
			/*
			** nothing to be done---will exit at top of loop
			** with EOF
			*/
			break;
		default:
			fprintf(stderr, "error in PFB file\n");
			exit(1);
		}
	}
	if (last_type != ASCII) {
		fputc('\n', psg->file);
	}
}
